# -*- coding: utf-8 -*-

import math

from .utils import secure_random, load_word_list
from .constants import *


__all__ = [
    "chars",
    "words"
]


def chars(charset=None, length=None, min_entropy=None):
    """Generates a character-based password. If the length parameter is supplied, the min_entropy parameter
    is ignored (i.e. either a length or a minimum entropy is required, but not both). If no length or
    min_entropy parameters are supplied, a default password length is chosen (see
    constants.DEFAULT_CHAR_PASSWORD_LENGTH).

    Args:
        charset: The character set to use from which to source characters. If not specified, it defaults to
            the character set with alphanumeric and special characters.
        length: The desired length of the password.
        min_entropy: The desired minimum entropy of the password, based on the given charset.

    Returns:
        A string containing the generated password.
    """
    if charset not in PASSWORD_CHARSETS:
        raise ValueError("Unrecognised charset: %s" % charset)
    else:
        charset_chars = list(PASSWORD_CHARSETS[charset])
        charset_size = len(charset_chars)

    password = ""

    if length is None and min_entropy is None:
        length = DEFAULT_CHAR_PASSWORD_LENGTH

    if length is not None:
        for i in range(length):
            password += charset_chars[secure_random(charset_size)]

    else:
        # work backwards from the entropy
        entropy_per_char = math.log(charset_size, 2.0)
        # round up on the number of characters
        min_chars = int(math.ceil(min_entropy / entropy_per_char))
        for i in range(min_chars):
            password += charset_chars[secure_random(charset_size)]

    return password


def select_random_words(word_list, count, starting_letters=None):
    if starting_letters is None:
        total_words = len(word_list)
        return [word_list[secure_random(total_words)] for i in range(count)]
    else:
        # assume word_list is a dictionary
        result = []
        for i in range(count):
            ch = starting_letters[i]
            result.append(word_list[ch][secure_random(len(word_list[ch]))])
        return result


def words(dict_set=None, separator=None, word_count=None, min_entropy=None, starting_letters=None):
    """Generates a word-based password from the given dictionary. If the word_count parameter is supplied,
    the min_entropy parameter is ignored (i.e. either a word count or minimum entropy is required, but not
    both). If no length or min_entropy parameters are supplied, a default word count is chosen (see
    constants.DEFAULT_WORD_PASSWORD_WORDS).

    Args:
        dict_set: The word list/dictionary from which to generate a password. Defaults to the built-in word list.
        separator: The separator to use between words.
        word_count: The number of words to use to build the password.
        min_entropy: The desired minimum entropy of the password, based on the given dictionary.
        starting_letters: A string containing the desired starting letters of the generated words. If word_count or
            min_entropy are specified, the number of letters in this string must exactly match the intended number
            of words that will be generated.

    Returns:
        A string containing the generated password.
    """
    if dict_set is None:
        dict_set = load_word_list()

    word_list = list(dict_set)
    word_list_size = len(word_list)

    # words categorised by alphabetical character
    categorised_words = dict()
    if starting_letters is not None:
        # make sure it's lowercase
        starting_letters = starting_letters.lower()
        for word in dict_set:
            ch = word[0]
            if ch in categorised_words:
                categorised_words[ch].append(word)
            else:
                categorised_words[ch] = [word]

        # check that all of the required starting letters are represented
        for ch in starting_letters:
            if not (ch in categorised_words):
                raise ValueError("Dictionary does not contain any words beginning with \"%s\"" % ch)

    password_words = []

    if word_count is None and min_entropy is None:
        word_count = DEFAULT_WORD_PASSWORD_WORDS if starting_letters is None else len(starting_letters)

    if separator is None:
        separator = DEFAULT_WORD_SEPARATOR

    if word_count is not None:
        if starting_letters is None:
            password_words.extend(select_random_words(word_list, word_count))

        else:
            if len(starting_letters) < word_count:
                raise ValueError((
                    "Please supply at least %d starting letters to meet the minimum word count " +
                    "requirement"
                ) % word_count)

            password_words.extend(
                select_random_words(
                    categorised_words,
                    word_count,
                    starting_letters=starting_letters
                )
            )

    else:
        entropy_per_word = math.log(word_list_size, 2.0)
        min_words = int(math.ceil(min_entropy / entropy_per_word))

        if starting_letters is None:
            password_words.extend(select_random_words(word_list, min_words))

        else:
            if len(starting_letters) < min_words:
                raise ValueError((
                    "Please supply at least %d starting letters to meet the minimum word count " +
                    "requirement"
                ) % min_words)

            password_words.extend(
                select_random_words(
                    categorised_words,
                    min_words,
                    starting_letters=starting_letters
                )
            )

    return separator.join(password_words)



# -*- coding: utf-8 -*-

from functools import reduce
from operator import mul
from string import ascii_lowercase
from io import open
import time
import math
import pkg_resources
import os
import struct

from .constants import *


__all__ = [
    "clean_word_list",
    "permutations",
    "calculate_entropy",
    "load_word_list",
    "secure_random",
    "secure_random_quality"
]


def clean_word_list(input_path, output_path, encoding=None, min_word_len=None):
    """Cleans the given word list, ensuring no punctuation or capitalisation or duplicate words. Word lists
    must be plain text files, with one word per line, and sorted alphabetically.

    Args:
        input_path: The path to the input word list file.
        output_path: The path to where to write the output, filtered word list.
        encoding: The encoding to use when attempting to read the file (default: platform-dependent).
        min_word_len: The minimum length of words to include. Defaults to constants.DEFAULT_MIN_WORD_LEN.

    Returns:
        A dictionary containing statistics about the clean operation.
    """
    if min_word_len is None:
        min_word_len = DEFAULT_MIN_WORD_LEN
    word_list = set()
    start_time = time.time()
    words_read = 0
    with open(input_path, "rt", encoding=encoding) as input_file:
        for line in input_file:
            word = line.strip().lower()
            if len(word) > 0:
                words_read += 1

                # strip possessives
                if word.endswith("'s"):
                    word = word[:-2]

                # strip non-alphabetical characters
                stripped_word = ""
                for c in word:
                    if c in ascii_lowercase:
                        stripped_word += c

                if stripped_word and len(stripped_word) >= min_word_len:
                    word_list.add(stripped_word)

    with open(output_path, "wt", encoding=encoding) as output_file:
        for word in sorted(list(word_list)):
            output_file.write("%s\n" % word)

    end_time = time.time()
    return {
        "time": end_time - start_time,
        "words_read": words_read,
        "words_written": len(word_list)
    }


def permutations(n, k):
    """Calculates the number of ordered k-permutations of n.

    Args:
        n: The total size of the set.
        k: The size of each sub-set.

    Returns:
        The number of possible uniquely ordered subsets of length k that can be generated from the set of size n.
    """
    return reduce(mul, range(n - k + 1, n + 1)) if 0 <= k <= n else 0


def calculate_entropy(password, dict_set=None):
    """Utility to calculate the entropy of a password (in bits) based on the detected charset (as from the
    perspective of a prospective attacker).

    Args:
        password: The source password to use in calculation.
        dict_set: The set of words in our dictionary/word list.

    Returns:
        A dictionary containing the entropies of the password based on different attacker dictionaries.
    """
    password_letters = set(password)
    password_len = len(password)
    entropy = dict()

    # find the charsets in which we'll find this password
    for charset_name, charset in PASSWORD_CHARSETS.items():
        if password_letters.issubset(charset):
            entropy[charset_name] = math.log(1.0*len(charset), 2.0) * password_len

    if dict_set is not None:
        # we assume our dictionary words are all lowercase, and that our separator is used
        if password_letters.issubset(PASSWORD_CHARSETS[PC_ALPHA_LOWER_SEP]):
            # detect the separator
            sep = None
            for c in password_letters:
                if c in separators:
                    sep = c
                    break

            # split the words by separator
            words = password.split(sep) if sep is not None else [password]
            # only if the words are unique to each other
            if len(words) == len(set(words)):
                all_words_found = True
                for word in words:
                    if word not in dict_set:
                        all_words_found = False

                # only if all of the words in the password are in our specific dictionary
                if all_words_found:
                    entropy[PC_DICT] = math.log(permutations(len(dict_set), len(words)), 2.0)

    return entropy


def load_word_list(filename=None, resource=None, encoding=None):
    """Loads a word list from the given filename or resource.

    Args:
        filename: If specified, loads the word list from this file system path.
        resource: If no filename is specified, this is loaded relative to the passwdgen package path. If no resource
            is specified, the default word list is used.
        encoding: The encoding to use when reading the file (default: OS-dependent).

    Returns:
        A set containing the entire list of non-zero-length words in the word list.
    """
    words = set()

    if filename is None:
        filename = pkg_resources.resource_filename("passwdgen", resource or DEFAULT_WORD_LIST)

    with open(filename, "rt", encoding=encoding) as input_file:
        for line in input_file:
            word = line.strip()
            if len(word) > 0:
                words.add(word)

    if len(words) < MIN_DICT_SIZE:
        raise ValueError("Dictionary is too small. Valid dictionaries must contain at least %d unique words." % MIN_DICT_SIZE)

    return words


def secure_random(a, b=None):
    """Generates integers in the most secure manner possible provided by the operating system. On POSIX machines,
    this will use /dev/urandom. On Windows machines, this will use CryptGenRandom().

    Args:
        a: If b is supplied, this should be the minimum value of the randomly generated number (inclusive). If b
            is not supplied, this should be the maximum value of the randomly generated number (exclusive), and the
            minimum will be 0.
        b: If supplied, this must provide the maximum value of the randomly generated number (exclusive).

    Returns:
        A random integer i, with a <= i < b if b is supplied, otherwise 0 <= i < a.
    """
    if (a < 0) or ((b is not None) and (b < 0)):
        raise ValueError("Both a and b need to be integers >= 0 for secure random number generation")
    if (b is not None) and (b <= a):
        raise ValueError("For secure random number generation, b must be < a")

    (random_val, ) = struct.unpack("Q", os.urandom(8))
    return (random_val % int(a)) if b is None else (int(a) + (random_val % int(b - a)))


def secure_random_quality(sample_size=1000000):
    """Attempts to estimate the quality of the secure random number generator of the operating system.

    Args:
        sample_size: The number of random samples to generate and test.

    Returns:
        A dictionary containing some statistics about the random number generator.
    """
    start_time = time.time()

    total = 0
    counts = dict()
    for i in range(sample_size):
        rnd_val = secure_random(101)
        total += rnd_val
        counts[rnd_val] = (counts[rnd_val] + 1) if rnd_val in counts else 1

    mean = float(total) / float(sample_size)
    # calculate the variance
    variance = 0.0
    for val, count in counts.items():
        variance += ((val - mean) ** 2.0) * count
    variance /= float(sample_size) - 1.0
    # ensure variance is positive (sometimes zeros can be negative with floating point numbers)
    if variance < 0.0:
        variance *= -1.0
    stddev = math.sqrt(variance)
    expected_variance = ((101.0 ** 2.0) - 1.0) / 12.0
    expected_stddev = math.sqrt(expected_variance)

    end_time = time.time()
    return {
        "mean": mean,
        "expected_mean": 50.0,
        "mean_diff": 100.0 * (abs(50.0 - mean) / 50.0),
        "stddev": stddev,
        "expected_stddev": expected_stddev,
        "stddev_diff": 100.0 * (abs(expected_stddev - stddev) / expected_stddev),
        "time": end_time - start_time
    }



#!/usr/bin/env python

from setuptools import setup

INSTALL_REQUIREMENTS = [
    "pyperclip==1.7.0"
]

setup(
    name="passwdgen",
    version="0.3.0",
    description="A password generation utility",
    author="Thane Thomson",
    author_email="connect@thanethomson.com",
    url="https://github.com/thanethomson/passwdgen",
    install_requires=INSTALL_REQUIREMENTS,
    entry_points={
        "console_scripts": [
            "passwdgen = passwdgen.cmdline:main"
        ]
    },
    packages=["passwdgen"],
    package_data={
        "passwdgen": [
            "data/*.txt"
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Utilities",
    ]
)



# -*- coding: utf-8 -*-

import string

__all__ = [
    "PC_ALPHA_LOWER",
    "PC_ALPHA_UPPER",
    "PC_ALPHA",
    "PC_ALPHA_NUMERIC",
    "PC_NUMERIC",
    "PC_ALPHA_LOWER_SEP",
    "PC_ALPHA_UPPER_SEP",
    "PC_SPECIAL",
    "PC_DICT",
    "PASSWORD_CHARSETS",
    "separators",
    "SEP_NONE",
    "SEP_DASH",
    "SEP_UNDERSCORE",
    "SEP_PERIOD",
    "SEP_SPACE",
    "SEP_COMMA",
    "SEP_SEMICOLON",
    "SEP_COLON",
    "PASSWORD_SEPARATORS",
    "PASSWORD_SEPARATOR_IDS",
    "PASSWORD_CHARSET_NAMES",
    "PASSWORD_CHARSET_IDS",
    "LONGEST_CHARSET_NAME_LEN",
    "DEFAULT_WORD_LIST",
    "DEFAULT_CHAR_PASSWORD_LENGTH",
    "DEFAULT_WORD_PASSWORD_WORDS",
    "DEFAULT_MIN_WORD_LEN",
    "MIN_DICT_SIZE",
    "DEFAULT_WORD_SEPARATOR"
]

PC_ALPHA_LOWER = "alpha-lower"
PC_ALPHA_UPPER = "alpha-upper"
PC_ALPHA = "alpha"
PC_ALPHA_NUMERIC = "alpha-numeric"
PC_ALPHA_NUMERIC_SPACED = "alpha-numeric-spaced"
PC_NUMERIC = "numeric"
PC_ALPHA_LOWER_SEP = "alpha-lower-sep"
PC_ALPHA_UPPER_SEP = "alpha-upper-sep"
PC_SPECIAL = "special"
PC_DICT = "dict"

separators = "-_. ,;:"

# Use special characters from AWS password requirements:
# https://docs.aws.amazon.com/IAM/latest/UserGuide/id_credentials_passwords_account-policy.html
special_chars = "!@#$%^&*()_+-=[]{}|'"

SEP_NONE = "none"
SEP_DASH = "dash"
SEP_UNDERSCORE = "und"
SEP_PERIOD = "period"
SEP_SPACE = "space"
SEP_COMMA = "comma"
SEP_SEMICOLON = "semi"
SEP_COLON = "colon"

PASSWORD_SEPARATORS = {
    SEP_NONE: "",
    SEP_DASH: "-",
    SEP_UNDERSCORE: "_",
    SEP_PERIOD: ".",
    SEP_SPACE: " ",
    SEP_COMMA: ",",
    SEP_SEMICOLON: ";",
    SEP_COLON: ":"
}

PASSWORD_SEPARATOR_IDS = [
    SEP_NONE,
    SEP_DASH,
    SEP_UNDERSCORE,
    SEP_PERIOD,
    SEP_SPACE,
    SEP_COMMA,
    SEP_SEMICOLON,
    SEP_COLON
]

PASSWORD_CHARSETS = {
    PC_ALPHA_LOWER: set(string.ascii_lowercase),
    PC_ALPHA_UPPER: set(string.ascii_uppercase),
    PC_ALPHA: set(string.ascii_letters),
    PC_ALPHA_NUMERIC: set(string.ascii_letters + string.digits),
    PC_ALPHA_NUMERIC_SPACED: set(string.ascii_letters + string.digits + " "),
    PC_ALPHA_LOWER_SEP: set(string.ascii_lowercase + separators),
    PC_ALPHA_UPPER_SEP: set(string.ascii_uppercase + separators),
    PC_SPECIAL: set(string.ascii_letters + string.digits + special_chars),
    PC_NUMERIC: set(string.digits)
}

PASSWORD_CHARSET_NAMES = (
    (PC_ALPHA_LOWER, "Alphabetical, lowercase (a-z)"),
    (PC_ALPHA_UPPER, "Alphabetical, uppercase (A-Z)"),
    (PC_ALPHA, "Alphabetical (a-z, A-Z)"),
    (PC_ALPHA_NUMERIC, "Alphabetical and numeric (a-z, A-Z, 0-9)"),
    (PC_ALPHA_NUMERIC_SPACED, "Alphabetical, numeric, with spaces (a-z, A-Z, 0-9, space)"),
    (PC_NUMERIC, "Numeric (0-9)"),
    (PC_ALPHA_LOWER_SEP, "Alphabetical, lowercase, with separator (a-z, separator)"),
    (PC_ALPHA_UPPER_SEP, "Alphabetical, uppercase, with separator (A-Z, separator)"),
    (PC_SPECIAL, "Alphabetical, numeric and special characters (a-z, A-Z, 0-9, punctuation)"),
    (PC_DICT, "Dictionary")
)

PASSWORD_CHARSET_IDS = [id for id, _ in PASSWORD_CHARSET_NAMES]
LONGEST_CHARSET_NAME_LEN = max([len(name) for _, name in list(PASSWORD_CHARSET_NAMES)])

DEFAULT_CHARSET = PC_SPECIAL
DEFAULT_WORD_LIST = "data/default-word-list.txt"
DEFAULT_CHAR_PASSWORD_LENGTH = 12
DEFAULT_WORD_PASSWORD_WORDS = 4
DEFAULT_WORD_SEPARATOR = "-"
DEFAULT_MIN_WORD_LEN = 3

# minimum number of words in a dictionary
MIN_DICT_SIZE = 100



# -*- coding: utf-8 -*-

import sys
from getpass import getpass
import argparse
import pkg_resources
import pyperclip

from .generator import *
from .utils import *
from .constants import *


def show_password_entropy(passwd, word_list):
    """Displays the password entropy calculation results."""
    entropy = calculate_entropy(passwd, dict_set=word_list)
    print("\nPassword length: %d characters" % len(passwd))
    print("\nEntropy")
    print("-------")
    for charset, charset_name in list(PASSWORD_CHARSET_NAMES):
        print(("{:<%d}" % LONGEST_CHARSET_NAME_LEN).format(charset_name) + " : " +
              (("%.6f" % entropy[charset]) if charset in entropy else "not in character set"))
    print("")


def main():
    """Main routine for handling command line functionality for passwdgen."""

    version = pkg_resources.require("passwdgen")[0].version
    parser = argparse.ArgumentParser(description="A password generation utility (v%s)." % version)
    subparsers = parser.add_subparsers(help="The command to execute.", dest="command")

    parser_info = subparsers.add_parser(
        "info",
        help=(
            "Compute information about a password. If passwdgen has input piped into it via stdin, that " +
            "will be interpreted as the password."
        )
    )
    parser_info.add_argument(
        "-d", "--dictionary",
        default=None,
        help="Path to the dictionary file to use. This must be a plain text file with one word per line."
    )
    parser_info.add_argument(
        "-e", "--encoding",
        default=None,
        help=(
            "The encoding to use when read/writing input/output files. " +
            "(See https://docs.python.org/2/library/codecs.html#standard-encodings)"
        )
    )

    parser_generate = subparsers.add_parser(
        "generate",
        help="Generate password(s)."
    )
    parser_generate.add_argument(
        "-c", "--clipboard",
        action="store_true",
        help=(
            "Copy the generated password to the clipboard (only for when generating a single password) instead of "+
            "writing the password to stdout"
        )
    )
    parser_generate.add_argument(
        "-d", "--dictionary",
        default=None,
        help="Path to the dictionary file to use. This must be a plain text file with one word per line."
    )
    parser_generate.add_argument(
        "-e", "--encoding",
        default=None,
        help=(
            "The encoding to use when read/writing input/output files. " +
            "(See https://docs.python.org/2/library/codecs.html#standard-encodings)"
        )
    )
    parser_generate.add_argument(
        "-i", "--info",
        action="store_true",
        help="Additionally display information about the generated password, including password entropy."
    )
    parser_generate.add_argument(
        "-l", "--length",
        type=int,
        default=None,
        help=(
            "The default number of characters or words to generate, depending on which kind of password " +
            "is being generated (a character- or dictionary-based one). Defaults: %d characters or %d words."
        ) % (DEFAULT_CHAR_PASSWORD_LENGTH, DEFAULT_WORD_PASSWORD_WORDS)
    )
    parser_generate.add_argument(
        "-m", "--min-entropy",
        default=None,
        type=int,
        help="The minimum entropy of the required password (optional). If length is specified, this will be ignored."
    )
    parser_generate.add_argument(
        "-s", "--separator",
        choices=PASSWORD_SEPARATOR_IDS,
        default=SEP_DASH,
        help=(
            "The separator to use when generating passwords from dictionaries (default=%s)."
        ) % SEP_DASH
    )
    parser_generate.add_argument(
        "--starting-letters",
        default=None,
        help=(
            "The letters to use as initials for the generated words."
        )
    )
    parser_generate.add_argument(
        "-t", "--charset",
        choices=PASSWORD_CHARSET_IDS,
        default=PC_DICT,
        help=(
                 "Which character set/approach to use when generating the password (default=\"%s\"). See the " +
                 "README.md file at https://github.com/thanethomson/passwdgen for more details."
             ) % PC_DICT
    )

    parser_rng = subparsers.add_parser(
        "rng",
        help="Test the quality of the operating system's random number generator."
    )
    parser_rng.add_argument(
        "-s", "--sample-size",
        type=int,
        default=1000000,
        help="Define the sample size to test with (default = 1,000,000)."
    )

    subparsers.add_parser(
        "version",
        help="Display the version of passwdgen and exit."
    )

    parser_wordlist = subparsers.add_parser(
        "wordlist",
        help="Utilities relating to word list manipulation."
    )
    subparsers_wordlist = parser_wordlist.add_subparsers(dest="wordlist_subcommand")

    parser_wordlist_clean = subparsers_wordlist.add_parser(
        "clean",
        help="Cleans up a given word list, stripping punctuation, digits and whitespace."
    )
    parser_wordlist_clean.add_argument(
        "input_file",
        help="The input text file, one word per line, to be cleaned."
    )
    parser_wordlist_clean.add_argument(
        "output_file",
        help="The output file into which to write the cleaned word list."
    )
    parser_wordlist_clean.add_argument(
        "-e", "--encoding",
        default=None,
        help=(
            "The encoding to use when read/writing input/output files. " +
            "(See https://docs.python.org/2/library/codecs.html#standard-encodings)"
        )
    )

    args = parser.parse_args()

    if args.command == "version":
        print("passwdgen v%s" % version)

    elif args.command == "info":
        if sys.stdin.isatty():
            passwd = getpass("Please enter the password to check: ")
        else:
            # if the input's been piped in
            passwd = sys.stdin.read()
            # strip off the single trailing newline
            if passwd.endswith("\n"):
                passwd = passwd[:-1]

        word_list = load_word_list(filename=args.dictionary, encoding=args.encoding)
        show_password_entropy(passwd, word_list)

    elif args.command == "rng":
        print("Testing OS RNG. Attempting to generate %d samples between 0 and 100 (inclusive). Please wait..." % args.sample_size)
        result = secure_random_quality(args.sample_size)
        print("\nStatistics")
        print("----------")
        print("Mean               : %.6f (should approach %.3f as the sample size increases; %.3f%% difference)" % (
            result['mean'],
            result['expected_mean'],
            result['mean_diff']
        ))
        print("Standard deviation : %.6f (should be as close to %.6f as possible; %.3f%% difference)" % (
            result['stddev'],
            result['expected_stddev'],
            result['stddev_diff']
        ))
        print("Time taken         : %.3f seconds\n" % result['time'])

    elif args.command == "generate":
        try:
            word_list = load_word_list(filename=args.dictionary, encoding=args.encoding)

            # dictionary-based password generation
            if args.charset == PC_DICT:
                # load our dictionary
                passwd = words(
                    word_list,
                    separator=PASSWORD_SEPARATORS[args.separator],
                    word_count=args.length,
                    min_entropy=args.min_entropy,
                    starting_letters=args.starting_letters
                )
            else:
                passwd = chars(
                    args.charset,
                    length=args.length,
                    min_entropy=args.min_entropy
                )

            if args.clipboard:
                pyperclip.copy(passwd)
                print("Password copied to clipboard.")
            else:
                print(passwd)

            if args.info:
                show_password_entropy(passwd, word_list)

        except ValueError as e:
            print("Error: %s" % e)

    elif args.command == "wordlist":
        if args.wordlist_subcommand == "clean":
            print("Attempting to clean word list: %s" % args.input_file)
            result = clean_word_list(
                args.input_file,
                args.output_file,
                encoding=args.encoding
            )
            print("Cleaned file in %.3f seconds. Read %d words, wrote %d." % (
                result["time"],
                result["words_read"],
                result["words_written"]
            ))
